
#define	ENTRY(name) \
	.globl name; \
	.align 16; \
	name:

	.macro PUSH_ERROR
	sub $4, %esp		/* space for error code */
	.endm
	
	.macro PUSH_TRAP_EBP trapno
	sub $4, %esp		/* space for trap number */
	push %ebp
	mov $\trapno, %ebp
	mov %ebp, 4(%esp)	/* save trap number on stack */
	.endm

	.macro PUSH_REGS
	push %edi
	push %esi
	push %edx
	push %ecx
	push %ebx
	push %eax
	.endm

	.macro POP_REGS
	pop %eax
	pop %ebx
	pop %ecx
	pop %edx
	pop %esi
	pop %edi
	pop %ebp
	.endm

	.macro RETURN
	add $8, %esp		/* remove error code & trap number */
	iret			/* jump back */
	.endm	

	.macro DO_TRAP trapno func
	PUSH_TRAP_EBP \trapno
	PUSH_REGS
	call \func
	POP_REGS
	RETURN
	.endm

/* ------------------------------------------------------------------ */

	.code32
	.text

/* handled by emu */

ENTRY(boot)
	lea stack_high, %esp	/* setup stack */
	sub $56, %esp		/* sizeof(struct regs_32) */
	call do_boot
	POP_REGS
	RETURN

ENTRY(division_by_zero)
	PUSH_ERROR
	DO_TRAP 0 do_guest_forward

ENTRY(illegal_instruction)
	PUSH_ERROR
	DO_TRAP 3 do_illegal_instruction

ENTRY(general_protection)
	DO_TRAP 13 do_general_protection

ENTRY(page_fault)
	DO_TRAP 14 do_page_fault

ENTRY(event_callback)
	PUSH_ERROR
	DO_TRAP -1 do_event_callback

ENTRY(xen_hypercall)
	PUSH_ERROR
	PUSH_TRAP_EBP -1
	PUSH_REGS
	call do_hypercall
	POP_REGS
	add $8, %esp		/* remove error code & trap number */
	cmp $-1, -4(%esp)
	jne 1f
	out %al, $0xe0          /* let userspace handle it */
1:
	iret

ENTRY(debug_int3)
	PUSH_ERROR
	DO_TRAP -1 do_debug_int3

/* bounced to userspace */
	
ENTRY(int_unknown)
	out %al, $0xef
	jmp loop

loop:
	hlt
	jmp loop

/* helpers */

ENTRY(memcpy_pf)
	mov 4(%esp),%edi
	mov 8(%esp),%esi
	mov 12(%esp),%ecx
	cld
1:	rep movsb
	xor %eax,%eax
8:	ret

	.section .exfix, "ax"
9:	mov $-1, %eax
	jmp 8b
	.previous

	.section .extab, "a"
	.align 4
	.long 1b,9b
	.previous

ENTRY(copy32_pf)
	mov 4(%esp),%edi
	mov 8(%esp),%esi
1:	mov (%esi),%eax
2:	mov %eax,(%edi)
	xor %eax,%eax
8:	ret

	.section .exfix, "ax"
9:	mov $-1, %eax
	jmp 8b
	.previous

	.section .extab, "a"
	.align 4
	.long 1b,9b
	.long 2b,9b
	.previous

ENTRY(copy64_pf)
	mov 4(%esp),%edi
	mov 8(%esp),%esi
1:	mov (%esi), %ebx
2:	mov 4(%esi), %ecx
3:	mov (%edi), %eax
4:	mov 4(%edi), %edx
5:	cmpxchg8b (%edi)
	xor %eax,%eax
8:	ret

	.section .exfix, "ax"
9:	mov $-1, %eax
	jmp 8b
	.previous

	.section .extab, "a"
	.align 4
	.long 1b,9b
	.long 2b,9b
	.long 3b,9b
	.long 4b,9b
	.previous

ENTRY(instructions)
	.byte  0xee, 0x89, 0xca, 0xec,  0x8d, 0x70, 0x7e, 0x89
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop

/* data section */

	.data
	.globl stack_low, stack_high
stack_low:
	.fill 4096,4,0
stack_high:

/* boot page tables */

#define pageflags 0x063 /* preset, rw, accessed, dirty */
#define largepage 0x080 /* pse */

	.globl emu_pgd_32,emu_pgd_pae,emu_pmd_pae

	.align 4096
emu_pgd_32:
	.fill 1020,4,0
	.long pageflags + largepage
	.fill 3,4,0

	.align 4096
emu_pgd_pae:
	.fill 3,8,0
	.long emu_pmd_pae - 0xff000000 + 0x001
	.long 0
	.fill 508,8,0

	.align 4096
emu_pmd_pae:
	.fill 504,8,0
	.quad pageflags + largepage
	.fill 7,8,0
